import snoop


def empty_decorator(function):
    return function


@empty_decorator
@snoop.snoop(
    depth=2)  # Multi-line decorator for extra confusion!
@empty_decorator
@empty_decorator
def main():
    str(3)



'''
Usage:

    import snoop

    @snoop
    def your_function(x):
        ...

A log will be written to stderr showing the lines executed and variables
changed in the decorated function.

For more information, see https://github.com/alexmojaki/snoop
'''

import collections
import sys

from .configuration import Config, install
from .variables import Attrs, Exploding, Indices, Keys

__VersionInfo = collections.namedtuple('VersionInfo',
                                       ('major', 'minor', 'micro'))

try:
    from .version import __version__

except ImportError:  # pragma: no cover
    # version.py is auto-generated with the git tag when building
    __version__ = "???"

try:
    __version_info__ = __VersionInfo(*(map(int, __version__.split('.'))))
except ValueError:
    __version_info__ = None


config = Config()
snoop = config.snoop
pp = config.pp
spy = config.spy
install = staticmethod(install)

sys.modules['snoop'] = snoop  # make the module callable

# Add all the attributes to the 'module' so things can be imported normally
for key, value in list(globals().items()):
    if key in 'collections sys __VersionInfo key value config':
        # Avoid polluting the namespace
        continue

    setattr(snoop, key, value)



@snoop()
def foo(x):
    return x ** 2


def no_trace(_x):
    return 3


@snoop(depth=2)
def bar():
    str({no_trace(x) for x in (1, 2)})


@snoop()
def main():
    str({x for x in list(range(100))})
    str({x: x for x in list(range(100))})
    str({y: {x + y for x in list(range(3))} for y in list(range(3))})
    str({foo(x) for x in (1, 2)})
    str({no_trace(x) for x in (1, 2)})
    bar()



from collections import OrderedDict

import snoop


class WithSlots(object):
    __slots__ = ('x', 'y')

    def __init__(self):
        self.x = 3
        self.y = 4


@snoop.snoop(watch=(
        snoop.Keys('_d', exclude='c'),
        snoop.Attrs('_s'),
        snoop.Indices('_lst')[-3:],
        snoop.Attrs('_lst'),  # doesn't have attributes
))
def main():
    _d = OrderedDict([('a', 1), ('b', 2), ('c', 'ignore')])
    _s = WithSlots()
    _lst = list(range(1000))



import functools
import inspect
import os
import re
import sys
import threading
from collections import OrderedDict

# noinspection PyUnresolvedReferences
from cheap_repr import cheap_repr, find_repr_function, try_register_repr

from snoop.utils import (NO_BIRDSEYE, ArgDefaultDict,
                         _register_cheap_reprs, ensure_tuple,
                         is_comprehension_frame, iscoroutinefunction,
                         my_cheap_repr, no_args_decorator, pp_name_prefix,
                         truncate_list)

from .formatting import Event, Source
from .variables import BaseVariable, CommonVariable, Exploding

find_repr_function(str).maxparts = 100
find_repr_function(bytes).maxparts = 100
find_repr_function(object).maxparts = 100
find_repr_function(int).maxparts = 999999
cheap_repr.suppression_threshold = 999999


class FrameInfo(object):
    def __init__(self, frame):
        self.frame = frame
        self.local_reprs = {}
        self.last_line_no = frame.f_lineno
        self.comprehension_variables = OrderedDict()
        self.source = Source.for_frame(frame)
        code = frame.f_code
        self.is_generator = code.co_flags & inspect.CO_GENERATOR
        self.had_exception = False
        if is_comprehension_frame(frame):
            self.comprehension_type = (
                    re.match(r'<(\w+)comp>', code.co_name).group(1).title()
                    + u' comprehension'
            )
        else:
            self.comprehension_type = ''
        self.is_ipython_cell = (
                code.co_name == '<module>' and
                code.co_filename.startswith('<ipython-input-')
        )

    def update_variables(self, watch, watch_extras, event, whitelist):
        self.last_line_no = self.frame.f_lineno
        old_local_reprs = self.local_reprs
        self.local_reprs = OrderedDict(
            (source, my_cheap_repr(value))
            for source, value in
            self.get_local_reprs(watch, watch_extras, whitelist)
        )

        if self.comprehension_type:
            for name, value_repr in self.local_reprs.items():
                values = self.comprehension_variables.setdefault(name, [])
                if not values or values[-1] != value_repr:
                    values.append(value_repr)
                    values[:] = truncate_list(values, 11)
            if event in ('return', 'exception'):
                return [
                    (name, ', '.join(values))
                    for name, values in self.comprehension_variables.items()
                ]
            else:
                return []

        variables = []
        for name, value_repr in self.local_reprs.items():
            if name not in old_local_reprs or old_local_reprs[name] != value_repr:
                variables.append((name, value_repr))
        return variables

    def get_local_reprs(self, watch, watch_extras, whitelist):
        frame = self.frame
        code = frame.f_code
        var_names = [
            key for key in frame.f_locals
            if whitelist is None or key in whitelist
            if not key.startswith(pp_name_prefix)
        ]
        vars_order = code.co_varnames + code.co_cellvars + code.co_freevars + tuple(var_names)
        var_names.sort(key=vars_order.index)
        result_items = [
            (key, frame.f_locals[key])
             for key in var_names
        ]

        for variable in watch:
            result_items += sorted(variable.items(frame))

        for source, value in result_items:
            yield source, value
            for extra in watch_extras:
                try:
                    pair = extra(source, value)
                except Exception:
                    pass
                else:
                    if pair is not None:
                        assert len(pair) == 2, "Watch extra must return pair or None"
                        yield pair


thread_global = threading.local()
internal_directories = (os.path.dirname((lambda: 0).__code__.co_filename),)

try:
    # noinspection PyUnresolvedReferences
    import birdseye
except ImportError:
    pass
else:
    internal_directories += (os.path.dirname(birdseye.__file__),)


class TracerMeta(type):
    def __new__(mcs, *args, **kwargs):
        result = super(TracerMeta, mcs).__new__(mcs, *args, **kwargs)
        result.default = result()
        return result

    def __call__(cls, *args, **kwargs):
        if no_args_decorator(args, kwargs):
            return cls.default(args[0])
        else:
            return super(TracerMeta, cls).__call__(*args, **kwargs)

    def __enter__(self):
        return self.default.__enter__(context=1)

    def __exit__(self, *args):
        return self.default.__exit__(*args, context=1)


class Tracer(metaclass=TracerMeta):
    def __init__(
            self,
            watch=(),
            watch_explode=(),
            depth=1,
    ):
        self.watch = [
            v if isinstance(v, BaseVariable) else CommonVariable(v)
            for v in ensure_tuple(watch)
        ] + [
            v if isinstance(v, BaseVariable) else Exploding(v)
            for v in ensure_tuple(watch_explode)
        ]
        self.frame_infos = ArgDefaultDict(FrameInfo)
        self.depth = depth
        assert self.depth >= 1
        self.target_codes = set()
        self.target_frames = set()
        self.variable_whitelist = None

    def __call__(self, function):
        if iscoroutinefunction(function):
            raise NotImplementedError("coroutines are not supported, sorry!")

        self.target_codes.add(function.__code__)

        @functools.wraps(function)
        def simple_wrapper(*args, **kwargs):
            with self:
                return function(*args, **kwargs)

        @functools.wraps(function)
        def generator_wrapper(*args, **kwargs):
            gen = function(*args, **kwargs)
            method, incoming = gen.send, None
            while True:
                with self:
                    try:
                        outgoing = method(incoming)
                    except StopIteration:
                        return
                try:
                    method, incoming = gen.send, (yield outgoing)
                except Exception as e:
                    method, incoming = gen.throw, e

        if inspect.isgeneratorfunction(function):
            return generator_wrapper
        else:
            return simple_wrapper

    def __enter__(self, context=0):
        if not self.config.enabled:
            return

        self.config.thread_local.__dict__.setdefault('depth', -1)

        calling_frame = sys._getframe(context + 1)
        if not self._is_internal_frame(calling_frame):
            calling_frame.f_trace = self.trace
            self.target_frames.add(calling_frame)
            self.config.last_frame = calling_frame
            self.trace(calling_frame, 'enter', None)

        stack = thread_global.__dict__.setdefault('original_trace_functions', [])
        stack.append(sys.gettrace())
        sys.settrace(self.trace)

    def __exit__(self, exc_type, exc_value, exc_traceback, context=0):
        if not self.config.enabled:
            return

        previous_trace = thread_global.original_trace_functions.pop()
        sys.settrace(previous_trace)
        calling_frame = sys._getframe(context + 1)
        calling_frame.f_trace = previous_trace
        self.trace(calling_frame, 'exit', None)
        self.target_frames.discard(calling_frame)
        self.frame_infos.pop(calling_frame, None)

    def _is_internal_frame(self, frame):
        return frame.f_code.co_filename.startswith(internal_directories)

    def _is_traced_frame(self, frame):
        return frame.f_code in self.target_codes or frame in self.target_frames

    def trace(self, frame, event, arg):
        if not self._is_traced_frame(frame):
            if (
                    self.depth == 1
                    or self._is_internal_frame(frame)
            ) and not is_comprehension_frame(frame):
                return None
            else:
                candidate = frame
                i = 0
                while True:
                    if is_comprehension_frame(candidate):
                        candidate = candidate.f_back
                        continue
                    i += 1
                    if self._is_traced_frame(candidate):
                        break
                    candidate = candidate.f_back
                    if i >= self.depth or candidate is None or self._is_internal_frame(candidate):
                        return None

        thread_local = self.config.thread_local
        frame_info = self.frame_infos[frame]
        if event in ('call', 'enter'):
            thread_local.depth += 1
        elif self.config.last_frame and self.config.last_frame is not frame:
            line_no = frame_info.last_line_no
            trace_event = Event(frame_info, event, arg, thread_local.depth, line_no=line_no)
            line = self.config.formatter.format_line_only(trace_event)
            self.config.write(line)

        if event == 'exception':
            frame_info.had_exception = True

        self.config.last_frame = frame

        trace_event = Event(frame_info, event, arg, thread_local.depth)
        if not (frame.f_code.co_name == '<genexpr>' and event not in ('return', 'exception')):
            trace_event.variables = frame_info.update_variables(
                self.watch,
                self.config.watch_extras,
                event,
                self.variable_whitelist,
            )

        if event in ('return', 'exit'):
            del self.frame_infos[frame]
            thread_local.depth -= 1

        formatted = self.config.formatter.format(trace_event)
        self.config.write(formatted)

        return self.trace

    @staticmethod
    def load_ipython_extension(ipython_shell):
        from snoop.ipython import SnoopMagics
        ipython_shell.register_magics(SnoopMagics)


class Spy(object):
    def __init__(self, config):
        self.config = config

    def __call__(self, *args, **kwargs):
        if NO_BIRDSEYE:
            raise Exception("birdseye doesn't support this version of Python")

        try:
            import birdseye
        except ImportError:
            raise Exception("You must install birdseye separately to use spy: pip install birdseye")

        # Decorator without parentheses
        if no_args_decorator(args, kwargs):
            return self._trace(args[0])

        # Decorator with parentheses and perhaps arguments
        def decorator(func):
            return self._trace(func, *args, **kwargs)

        return decorator

    def _trace(self, func, *args, **kwargs):
        # noinspection PyUnresolvedReferences
        from birdseye import eye

        traced = eye(func)
        traced = self.config.snoop(*args, **kwargs)(traced)

        _register_cheap_reprs()  # Override birdseye in case it's outdated

        @functools.wraps(func)
        def wrapper(*func_args, **func_kwargs):
            if self.config.enabled:
                final_func = traced
            else:
                final_func = func

            return final_func(*func_args, **func_kwargs)

        return wrapper




from birdseye import eye


@spy(depth=4)
def foo():
    x = pp.deep(lambda: 1 + 2)
    y = pp(3 + 4, 5 + 6)[0]
    return x + y


# Testing spy with and without parentheses
@spy
def bar():
    result = pp(4 + 6, 7 + 8)
    return result[0]


def main():
    for func in [foo, bar]:
        func()
        call_id = eye._last_call_id
        with eye.db.session_scope() as session:
            call = session.query(eye.db.Call).filter_by(id=call_id).one()
            assert call.result == '10'



from __future__ import division


@snoop(depth=3)
def main():
    x = 1
    y = 2
    pp(pp(x + 1) + max(*pp(y + 2, y + 3)))
    assert pp.deep(lambda: x + 1 + max(y + 2, y + 3)) == 7
    lst = list(range(30))
    pp.deep(lambda: list(
        list(a + b for a in [1, 2])
        for b in [3, 4]
    ) + lst)
    pp(dict.fromkeys(range(30), 4))
    pp.deep(lambda: BadRepr() and 1)
    pp.deep(lambda: 1 / 2)

    try:
        pp.deep(lambda: max(y + 2, (y + 3) / 0))
    except ZeroDivisionError:
        pass
    else:
        assert 0

    f = pp.deep(lambda: caller(lambda: 2 * 3))
    f()


class BadRepr(object):
    def __repr__(self):
        raise ValueError('bad')


def caller(f):
    f()
    return f



import snoop


def f4(x4):
    result4 = x4 * 2
    return result4


def f3(x3):
    result3 = f4(x3)
    return result3


def f2(x2):
    result2 = f3(x2)
    return result2


def main():
    str(3)
    with snoop.snoop(depth=3):
        result1 = f2(5)
    return result1



import ast
import inspect
import traceback
from copy import deepcopy
from threading import current_thread
from uuid import uuid4

from executing import future_flags, only

from snoop.formatting import Event, Source
from snoop.tracer import FrameInfo
from snoop.utils import (FormattedValue, builtins,
                         optional_numeric_label, pp_name_prefix)


class PP(object):
    def __init__(self, config):
        self.config = config

    def __call__(self, *args):
        if self.config.enabled:
            PPEvent(self, args, deep=False)

        if len(args) == 1:
            return args[0]
        else:
            return args

    def deep(self, arg):
        if not is_deep_arg(arg):
            raise TypeError("Argument must be a lambda without arguments")

        if self.config.enabled:
            return PPEvent(self, [arg], deep=True).returns

        return arg()


class PPEvent(object):
    def __init__(self, pp_object, args, deep):
        self.config = pp_object.config
        self.args = args
        depth = getattr(self.config.thread_local, 'depth', 0)
        frame = inspect.currentframe().f_back.f_back
        self.event = Event(FrameInfo(frame), 'log', None, depth)
        formatted = self.config.formatter.format_log(self.event)
        self.config.write(formatted)

        self.returns = None
        try:
            self.call = call = Source.executing(frame).node
            assert isinstance(call, ast.Call)
            assert len(args) == len(call.args)
        except Exception:
            if deep:
                self.returns = args[0] = args[0]()
            for i, arg in enumerate(args):
                self.write_placeholder(i, arg)
        else:
            if deep:
                call_arg = only(call.args)
                assert isinstance(call_arg, ast.Lambda), "You must pass a lambda DIRECTLY to pp.deep, not as a result of any other expression"
                self.returns = self.deep_pp(call_arg.body, frame)
            else:
                self.plain_pp(args, call.args)

    def write(self, source, value, depth=0):
        if depth == 0:
            value_string = self.config.pformat(value)
        else:
            try:
                value_string = repr(value)
            except Exception as e:
                exception_string = ''.join(
                    traceback.format_exception_only(type(e), e)
                ).strip()
                value_string = '<Exception in repr(): {}>'.format(exception_string)

        formatted = self.config.formatter.format_log_value(
            self.event, source, value_string, depth)
        self.config.write(formatted)

    def write_node(self, node, value, depth=0):
        source = self.event.source.get_text_with_indentation(node)
        self.write(source, value, depth=depth)

    def write_placeholder(self, i, arg):
        source = '<argument{}>'.format(optional_numeric_label(i, self.args))
        return self.write(source, arg)

    def plain_pp(self, args, call_args):
        for i, (call_arg, arg) in enumerate(zip(call_args, args)):
            try:
                self.write_node(call_arg, arg)
            except Exception:
                self.write_placeholder(i, arg)

    def deep_pp(self, call_arg, frame):
        stack = []
        thread = current_thread()

        def before_expr(tree_index):
            node = self.event.source.nodes[tree_index]
            if thread == current_thread():
                stack.append(node)
            return node

        before_expr.name = pp_name_prefix + 'before_' + uuid4().hex

        def after_expr(node, value):
            if thread == current_thread():
                assert node is stack.pop()

            try:
                ast.literal_eval(node)
                is_obvious = True
            except ValueError:
                is_obvious = (
                    isinstance(node, ast.Name)
                    and getattr(builtins, node.id, object()) == value
                )

            if not is_obvious:
                self.write_node(node, value, depth=node._depth - call_arg._depth)
            return value

        after_expr.name = pp_name_prefix + 'after_' + uuid4().hex

        new_node = deepcopy(call_arg)
        new_node = NodeVisitor(before_expr.name, after_expr.name).visit(new_node)
        expr = ast.Expression(new_node)
        ast.copy_location(expr, new_node)
        code = compile(
            expr,
            frame.f_code.co_filename,
            'eval',
            dont_inherit=True,
            flags=future_flags & frame.f_code.co_flags,
        )
        frame.f_globals[before_expr.name] = before_expr
        frame.f_globals[after_expr.name] = after_expr
        try:
            return eval(code, frame.f_globals, frame.f_locals)
        except Exception as e:
            if stack:
                last_node = stack[-1]
                self.write_node(
                    last_node,
                    DirectRepr('!!! {}!'.format(e.__class__.__name__)),
                    last_node._depth - call_arg._depth,
                )
            raise
        finally:
            frame.f_globals[before_expr.name] = lambda x: x
            frame.f_globals[after_expr.name] = lambda node, value: value


class DirectRepr(str):
    def __repr__(self):
        return self


try:
    getargspec = inspect.getfullargspec
except AttributeError:
    getargspec = inspect.getargspec


def is_deep_arg(x):
    return (
        inspect.isfunction(x)
        and x.__code__.co_name == '<lambda>'
        and not any(getargspec(x))
    )


class NodeVisitor(ast.NodeTransformer):
    """
    This does the AST modifications that call the hooks.
    """

    def __init__(self, before_name, after_name):
        self.before_name = before_name
        self.after_name = after_name

    def generic_visit(self, node):
        # type: (ast.AST) -> ast.AST
        if (isinstance(node, ast.expr) and
                not (hasattr(node, "ctx") and not isinstance(node.ctx, ast.Load)) and
                not isinstance(node, getattr(ast, 'Starred', ()))):
            return self.visit_expr(node)
        return super(NodeVisitor, self).generic_visit(node)

    def visit_expr(self, node):
        # type: (ast.expr) -> ast.Call
        """
        each expression e gets wrapped like this:
            _treetrace_hidden_after_expr(_treetrace_hidden_before_expr(_tree_index), e)

        where the _treetrace_* functions are the corresponding methods with the
        TreeTracerBase and traced_file arguments already filled in (see _trace_methods_dict)
        """

        before_marker = ast.Call(
            func=ast.Name(id=self.before_name,
                          ctx=ast.Load()),
            args=[ast.Num(node._tree_index)],
            keywords=[],
        )

        ast.copy_location(before_marker, node)

        if isinstance(node, FormattedValue):
            arg = node
        else:
            arg = super(NodeVisitor, self).generic_visit(node)

        after_marker = ast.Call(
            func=ast.Name(id=self.after_name,
                          ctx=ast.Load()),
            args=[
                before_marker,
                arg,
            ],
            keywords=[],
        )
        ast.copy_location(after_marker, node)
        ast.fix_missing_locations(after_marker)

        return after_marker



from threading import Thread

from snoop.configuration import Config

snoop = Config(columns='thread').snoop


@snoop
def foo():
    return 1


def run(name):
    thread = Thread(target=foo, name=name)
    thread.start()
    thread.join()


def main():
    run('short')
    run('longername')
    run('short')



def f2(a):
    def f3(a):
        x = 0
        x += 1

        def f4(_a):
            _y = x
            return 42

        return f4(a)

    return f3(a)


def main():
    with snoop(depth=4):
        result1 = f2(42)
    return result1



import snoop
import numpy as np
import pandas as pd


@snoop.snoop()
def main():
    arr = np.arange(10000)
    arr = arr.reshape((100, 100))
    df = pd.DataFrame(arr)
    series = df[0]
    return series



import snoop


@snoop
def foo(e):
    try:
        if e:
            raise TypeError('bad')
        return
    finally:
        pass


def main():
    foo(0)
    try:
        foo(1)
    except TypeError:
        pass


if __name__ == '__main__':
    main()



from __future__ import print_function

from snoop import Config

config = Config(color=True)
snoop = config.snoop
pp = config.pp


def foo():
    raise TypeError


@snoop(depth=2)
def main():
    try:
        (None
         or foo)()
    except:
        pass
    pp(
        [
            x + y
            for x, y in zip(range(1000, 1020), range(2000, 2020))
        ]
    )



from snoop.configuration import Config


def watch_type(source, value):
    return 'type({})'.format(source), type(value).__name__


def foo():
    x = 1
    y = [x]
    return y


def main():
    Config(watch_extras=watch_type).snoop(foo)()
    Config(replace_watch_extras=watch_type).snoop(foo)()



from snoop.configuration import Config

config = Config(enabled=False)

# This test leaves out spy for the NO_ASTTOKENS versions,
# but other versions can test it too

@config.snoop
def main():
    assert config.pp(1 + 2) == 3
    assert config.pp(1 + 3, 6) == (4, 6)



import ast

from IPython.core.magic import Magics, cell_magic, magics_class

from snoop import snoop


@magics_class
class SnoopMagics(Magics):
    @cell_magic
    def snoop(self, _line, cell):
        filename = self.shell.compile.cache(cell)
        code = self.shell.compile(cell, filename, 'exec')
        tracer = snoop()

        tracer.variable_whitelist = set()
        for node in ast.walk(ast.parse(cell)):
            if isinstance(node, ast.Name):
                name = node.id

                if isinstance(
                        self.shell.user_global_ns.get(name),
                        type(ast),
                ):
                    # hide modules
                    continue

                tracer.variable_whitelist.add(name)

        tracer.target_codes.add(code)
        with tracer:
            self.shell.ex(code)



from __future__ import print_function

import io
import os
import re
import sys
import traceback
from importlib import import_module
from tempfile import mkstemp
from threading import current_thread

import pytest
from cheap_repr import cheap_repr, register_repr
from cheap_repr.utils import safe_qualname
from littleutils import file_to_string, group_by_key_func, string_to_file

# Hide 3rd party pretty-printing modules
sys.modules['prettyprinter'] = {}
sys.modules['pprintpp'] = {}

from snoop import formatting, install, spy
from snoop.configuration import Config
from snoop.pp_module import is_deep_arg
from snoop.utils import (NO_BIRDSEYE, PYPY, needs_parentheses,
                         truncate_list, truncate_string)

formatting._get_filename = lambda _: "/path/to_file.py"

install()

tests_dir = os.path.dirname(__file__)
cheap_repr.raise_exceptions = True


@register_repr(type(cheap_repr))
def repr_function(func, _helper):
    return '<function %s at %#x>' % (
        safe_qualname(func), id(func))


@register_repr(type(sys))
def repr_module(module, _helper):
    return "<module '%s'>" % module.__name__


@register_repr(set)
def repr_set(x, helper):
    if not x:
        return repr(x)
    return helper.repr_iterable(x, '{', '}')


def sample_traceback():
    raw = u''.join(
        traceback.format_exception(*sys.exc_info())
    ).splitlines(True)
    tb = u''.join(
        line for line in raw
        if not line.strip().startswith("File")
    )
    sys.stderr.write(tb)


def assert_sample_output(module_name):
    module = import_module('tests.samples.' + module_name)
    old = sys.stderr

    out = io.StringIO()
    sys.stderr = out

    try:
        assert sys.gettrace() is None
        module.main()
        assert sys.gettrace() is None
    finally:
        sys.stderr = old

    time = '12:34:56.78'
    time_pattern = re.sub(r'\d', r'\\d', time)

    normalised = re.sub(time_pattern, time, out.getvalue()).strip()
    normalised = re.sub(r'0x\w+', '0xABC', normalised)
    normalised = normalised.replace('<genexpr>.<genexpr>', '<genexpr>')
    normalised = normalised.replace('<list_iterator', '<tupleiterator')
    normalised = normalised.replace('<listiterator', '<tupleiterator')
    normalised = normalised.replace('<tuple_iterator', '<tupleiterator')
    normalised = normalised.replace('<sequenceiterator', '<tupleiterator')
    normalised = normalised.replace(str(current_thread().ident), '123456789')

    result_filename = os.path.join(
        tests_dir,
        'sample_results',
        PYPY * 'pypy' + '.'.join(sys.version.split('.')[:2]),
        module_name + '.txt',
    )

    compare_to_file(normalised, result_filename)


def compare_to_file(text, filename):
    if os.environ.get('FIX_SNOOP_TESTS'):
        string_to_file(text, filename)
    else:
        expected_output = file_to_string(filename)
        assert text == expected_output


def generate_test_samples():
    samples_dir = os.path.join(tests_dir, 'samples')
    for filename in os.listdir(samples_dir):
        if filename.endswith('.pyc'):
            continue
        module_name = str(filename.split('.')[0])

        if module_name in '__init__ __pycache__':
            continue

        if PYPY or sys.version_info[:2] in ((3, 4), (3, 10)):
            if module_name in 'pandas_sample'.split():
                continue
        if NO_BIRDSEYE:
            if module_name in 'spy enabled'.split():
                continue

        if module_name in 'f_string' and sys.version_info[:2] < (3, 6):
            continue

        if module_name == 'color' and sys.version_info[:2] == (3, 12) and os.environ.get('CI'):
            continue

        yield module_name


@pytest.mark.parametrize("module_name", generate_test_samples())
@pytest.mark.order(1)
def test_sample(module_name):
    assert_sample_output(module_name)


@pytest.mark.order(2)  # Execute after all test_samples have run.
def test_compare_versions():
    out = [""]

    def prn(*args):
        out[0] += " ".join(map(str, args)) + "\n"

    samples_dir = os.path.join(tests_dir, "samples")
    results_dir = os.path.join(tests_dir, "sample_results")
    versions = sorted(os.listdir(results_dir))
    for filename in sorted(os.listdir(samples_dir)):
        if not filename.endswith(".py"):
            continue
        module_name = filename[:-3]
        if module_name == "__init__":
            continue

        doesnt_exist = "Doesn't exist:"

        def get_results(version):
            path = os.path.join(results_dir, version, module_name + ".txt")
            if os.path.exists(path):
                return file_to_string(path)
            else:
                return doesnt_exist

        grouped = group_by_key_func(versions, get_results)

        if len(grouped) > 1:
            prn(module_name)
            doesnt_exist_versions = grouped.pop(doesnt_exist, None)
            if doesnt_exist_versions:
                prn(doesnt_exist, ", ".join(doesnt_exist_versions))
            if len(grouped) > 1:
                prn("Differing versions:")
                for version_group in sorted(grouped.values()):
                    prn(", ".join(version_group))
            prn()

    compare_to_file(out[0], os.path.join(tests_dir, "version_differences.txt"))


def test_string_io():
    string_io = io.StringIO()
    config = Config(out=string_io)
    contents = u'stuff'
    config.write(contents)
    assert string_io.getvalue() == contents


def test_callable():
    string_io = io.StringIO()

    def write(msg):
        string_io.write(msg)

    string_io = io.StringIO()
    config = Config(out=write)
    contents = u'stuff'
    config.write(contents)
    assert string_io.getvalue() == contents


def test_file_output():
    _, path = mkstemp()
    config = Config(out=path)
    contents = u'stuff'
    config.write(contents)
    with open(path) as output_file:
        output = output_file.read()
    assert output == contents


def test_no_overwrite_by_default():
    _, path = mkstemp()
    with open(path, 'w') as output_file:
        output_file.write(u'lala')
    config = Config(str(path))
    config.write(u' doo be doo')
    with open(path) as output_file:
        output = output_file.read()
    assert output == u'lala doo be doo'


def test_overwrite():
    _, path = mkstemp()
    with open(path, 'w') as output_file:
        output_file.write(u'lala')

    config = Config(out=str(path), overwrite=True)
    config.write(u'doo be')
    config.write(u' doo')

    with open(path) as output_file:
        output = output_file.read()
    assert output == u'doo be doo'


def test_needs_parentheses():
    assert not needs_parentheses('x')
    assert not needs_parentheses('x.y')
    assert not needs_parentheses('x.y.z')
    assert not needs_parentheses('x.y.z[0]')
    assert not needs_parentheses('x.y.z[0]()')
    assert not needs_parentheses('x.y.z[0]()(3, 4 * 5)')
    assert not needs_parentheses('foo(x)')
    assert not needs_parentheses('foo(x+y)')
    assert not needs_parentheses('(x+y)')
    assert not needs_parentheses('[x+1 for x in ()]')
    assert needs_parentheses('x + y')
    assert needs_parentheses('x * y')
    assert needs_parentheses('x and y')
    assert needs_parentheses('x if z else y')


def test_truncate_string():
    max_length = 20
    for i in range(max_length * 2):
        string = i * 'a'
        truncated = truncate_string(string, max_length)
        if len(string) <= max_length:
            assert string == truncated
        else:
            assert truncated == 'aaaaaaaa...aaaaaaaaa'
            assert len(truncated) == max_length


def test_truncate_list():
    max_length = 5
    for i in range(max_length * 2):
        lst = i * ['a']
        truncated = truncate_list(lst, max_length)
        if len(lst) <= max_length:
            assert lst == truncated
        else:
            assert truncated == ['a', 'a', '...', 'a', 'a']
            assert len(truncated) == max_length


class Foo(object):
    def bar1(self):
        pass

    @staticmethod
    def bar2():
        pass


def test_is_deep_arg():
    assert is_deep_arg(lambda: 0)
    assert not is_deep_arg(lambda x: 0)
    assert not is_deep_arg(lambda x=None: 0)
    assert not is_deep_arg(lambda *x: 0)
    assert not is_deep_arg(lambda **x: 0)
    assert not is_deep_arg(test_is_deep_arg)
    assert not is_deep_arg(Foo)
    assert not is_deep_arg(Foo.bar1)
    assert not is_deep_arg(Foo().bar1)
    assert not is_deep_arg(Foo.bar2)
    assert not is_deep_arg(Foo().bar2)


def test_no_asttokens_spy():
    if PYPY:
        with pytest.raises(Exception, match="birdseye doesn't support this version of Python"):
            spy(test_is_deep_arg)



import snoop


def f4(x4):
    result4 = snoop.pp(x4 * 2)
    return result4


def f3(x3):
    result3 = f4(x3)
    return result3


def f2(x2):
    result2 = f3(x2)
    return result2


@snoop(depth=3)
def main():
    result1 = f2(8)
    return result1


if __name__ == '__main__':
    main()




from datetime import datetime
import sys

def test_misc():
    x = 1
    y = 2
    assert pp(pp(x) + y) == 3
    assert pp.deep(lambda: (x + y) + y) == 5

    with snoop:
        y = 3  # custom pformat is not used here

def test_larger_object():
    d = [
        'a long key to be pretty printed prettily',
        [
            'prettyprinter prints datetime with keyword arguments:',
            datetime.utcfromtimestamp(42),
        ]
    ]
    pp(d)

def main():
    sys.stderr.write(u'\nTest with custom pformat\n')
    snoop.install(pformat=lambda x: 'custom(' + repr(x) + ')')
    test_misc()
    test_larger_object()

    sys.stderr.write(u'\nTest without prettyprinter and without pprintpp\n')
    sys.modules['prettyprinter'] = {}
    sys.modules['pprintpp'] = {}
    snoop.install()
    test_larger_object()

    sys.stderr.write(u'\nTest with prettyprinter and without pprintpp\n')
    del sys.modules['prettyprinter']
    sys.modules['pprintpp'] = {}
    snoop.install()
    test_larger_object()

    sys.stderr.write(u'\nTest without prettyprinter and with pprintpp\n')
    sys.modules['prettyprinter'] = {}
    del sys.modules['pprintpp']
    snoop.install()
    test_larger_object()

    sys.stderr.write(u'\nTest with prettyprinter and with pprintpp\n')
    del sys.modules['prettyprinter']
    del sys.modules['pprintpp']
    snoop.install()
    test_larger_object()

    # clear 3rd party packages afterwards
    sys.modules['prettyprinter'] = {}
    sys.modules['pprintpp'] = {}
    snoop.install()




from snoop.configuration import Config

snoop = Config(prefix='ZZZ').snoop


class Baz(object):
    def __init__(self):
        self.x = 2

    @snoop(watch='self.x')
    def square(self):
        foo = 7
        self.x **= 2
        return self


def main():
    baz = Baz()
    baz.square()



import snoop


@snoop.snoop(depth=2)
def main():
    f2()


def f2():
    f3()


def f3():
    f4()


@snoop.snoop(depth=2)
def f4():
    f5()


def f5():
    pass


if __name__ == '__main__':
    main()



import inspect
import os
import sys
import threading
from collections.abc import Mapping, Sequence, Set
from io import open

import snoop as package
from snoop.formatting import DefaultFormatter
from snoop.pp_module import PP
from snoop.tracer import Spy, Tracer
from snoop.utils import QuerySet
from snoop.utils import builtins as builtins_module
from snoop.utils import ensure_tuple, is_pathlike, shitcode

try:
    # Enable ANSI escape codes in Windows 10
    import ctypes

    kernel32 = ctypes.windll.kernel32
    kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
    can_color = True
except Exception:
    can_color = os.name != 'nt'


def install(
        builtins=True,
        snoop="snoop",
        pp="pp",
        spy="spy",
        out=None,
        prefix='',
        columns='time',
        overwrite=False,
        color=None,
        enabled=True,
        watch_extras=(),
        replace_watch_extras=None,
        formatter_class=DefaultFormatter,
        pformat=None,
):
    """
    Configure output, enable or disable, and add names to builtins. Parameters:

    - builtins: set to False to not add any names to builtins,
        so importing will still be required.
    - snoop, pp, and spy: set to other strings
        to choose the names of these functions in builtins
    - `out`: determines the output destination. By default this is stderr. You can also pass:
        - A string or a `Path` object to write to a file at that location. By default this always will append to the file. Pass `overwrite=True` to clear the file initially.
        - Anything with a `write` method, e.g. `sys.stdout` or a file object.
        - Any callable with a single string argument, e.g. `logger.info`.
    - `color`: determines whether the output includes escape characters to display colored text in the console. If you see weird characters in your output, your console doesn't support colors, so pass `color=False`.
        - Code is syntax highlighted using [Pygments](http://pygments.org/), and this argument is passed as the style. You can choose a different color scheme by passing a string naming a style (see [this gallery](https://help.farbox.com/pygments.html)) or a style class. The default style is monokai.
        - By default this parameter is set to `out.isatty()`, which is usually true for stdout and stderr but will be false if they are redirected or piped. Pass `True` or a style if you want to force coloring.
        - To see colors in the PyCharm Run window, edit the Run Configuration and tick "Emulate terminal in output console".
    - `prefix`: Pass a string to start all snoop lines with that string so you can grep for them easily.
    - `columns`: This specifies the columns at the start of each output line. You can pass a string with the names of built in columns separated by spaces or commas. These are the available columns:
        - `time`: The current time. This is the only column by default.
        - `thread`: The name of the current thread.
        - `thread_ident`: The [identifier](https://docs.python.org/3/library/threading.html#threading.Thread.ident) of the current thread, in case thread names are not unique.
        - `file`: The filename (not the full path) of the current function.
        - `full_file`: The full path to the file (also shown anyway when the function is called).
        - `function`: The name of the current function.
        - `function_qualname`: The qualified name of the current function.

        If you want a custom column, please open an issue to tell me what you're interested in! In the meantime, you can pass a list, where the elements are either strings or callables. The callables should take one argument, which will be an `Event` object. It has attributes `frame`, `event`, and `arg`, as specified in [`sys.settrace()`](https://docs.python.org/3/library/sys.html#sys.settrace), and other attributes which may change.
    - `pformat`: set the pretty formatting function `pp` uses. Default is to use the first of `prettyprinter.pformat`, `pprintpp.pformat` and `pprint.pformat` that can be imported.
    """

    if builtins:
        setattr(builtins_module, snoop, package.snoop)
        setattr(builtins_module, pp, package.pp)
        setattr(builtins_module, spy, package.spy)
    config = Config(
        out=out,
        prefix=prefix,
        columns=columns,
        overwrite=overwrite,
        color=color,
        enabled=enabled,
        watch_extras=watch_extras,
        replace_watch_extras=replace_watch_extras,
        formatter_class=formatter_class,
        pformat=pformat,
    )
    package.snoop.config = config
    package.pp.config = config
    package.spy.config = config


class Config(object):
    """"
    If you need more control than the global `install` function, e.g. if you want to write to several different files in one process, you can create a `Config` object, e.g: `config = snoop.Config(out=filename)`. Then `config.snoop`, `config.pp` and `config.spy` will use that configuration rather than the global one.

    The arguments are the same as the arguments of `install()` relating to output configuration and `enabled`.
    """

    def __init__(
            self,
            out=None,
            prefix='',
            columns='time',
            overwrite=False,
            color=None,
            enabled=True,
            watch_extras=(),
            replace_watch_extras=None,
            formatter_class=DefaultFormatter,
            pformat=None,
    ):
        if can_color:
            if color is None:
                isatty = getattr(out or sys.stderr, 'isatty', lambda: False)
                color = bool(isatty())
        else:
            color = False

        self.write = get_write_function(out, overwrite)
        self.formatter = formatter_class(prefix, columns, color)
        self.enabled = enabled

        if pformat is None:
            try:
                from prettyprinter import pformat
            except Exception:
                try:
                    from pprintpp import pformat
                except Exception:
                    from pprint import pformat

        self.pformat = pformat

        self.pp = PP(self)

        class ConfiguredTracer(Tracer):
            config = self

        self.snoop = ConfiguredTracer
        self.spy = Spy(self)

        self.last_frame = None
        self.thread_local = threading.local()

        if replace_watch_extras is not None:
            self.watch_extras = ensure_tuple(replace_watch_extras)
        else:
            self.watch_extras = (len_shape_watch, dtype_watch) + ensure_tuple(watch_extras)


def len_shape_watch(source, value):
    try:
        shape = value.shape
    except Exception:
        pass
    else:
        if not inspect.ismethod(shape):
            return '{}.shape'.format(source), shape

    if isinstance(value, QuerySet):
        # Getting the length of a Django queryset evaluates it
        return None

    length = len(value)
    if (
            (isinstance(value, (str, bytes))
             and length < 50) or
            (isinstance(value, (Mapping, Set, Sequence))
             and length == 0)
    ):
        return None

    return 'len({})'.format(source), length


def dtype_watch(source, value):
    dtype = value.dtype
    if not inspect.ismethod(dtype):
        return '{}.dtype'.format(source), dtype


def get_write_function(output, overwrite):
    is_path = (
            isinstance(output, str)
            or is_pathlike(output)
    )
    if is_path:
        return FileWriter(output, overwrite).write
    elif callable(output):
        write = output
    else:
        def write(s):
            stream = output

            if stream is None:
                stream = sys.stderr

            try:
                stream.write(s)
            except UnicodeEncodeError:
                # God damn Python 2
                stream.write(shitcode(s))

            getattr(stream, "flush", lambda: None)()

    return write


class FileWriter(object):
    def __init__(self, path, overwrite):
        self.path = str(path)
        self.overwrite = overwrite

    def write(self, s):
        with open(self.path, 'w' if self.overwrite else 'a', encoding='utf-8') as f:
            f.write(s)
        self.overwrite = False



import snoop


def foo():
    raise TypeError('''
    very
    bad''')


def bar(*_):
    try:
        str(foo())
    except Exception:
        str(1)
        raise


@snoop(depth=3)
def main():
    try:
        bar()
    except:
        pass

    try:
        bob(
            1,
            2
        )
    except:
        pass

    try:
        (None
         or bob)(
            1,
            2
        )
    except:
        pass

    x = [[[2]]]

    try:
        str(x[1][0][0])
    except:
        pass

    try:
        str(x[0][1][0])
    except:
        pass

    try:
        str(x[0][0][1])
    except:
        pass


def bob(*_):
    pass


bob()
bob = None


if __name__ == '__main__':
    main()



import os
from snoop import snoop

@snoop()
def main():
    os.environ['DJANGO_SETTINGS_MODULE'] = 'tests.fake_django_settings'
    import django

    django.setup()
    from django.contrib.contenttypes.models import ContentType
    queryset = ContentType.objects.all()
    lst = [1, 2, 3]



from birdseye import eye

from snoop.configuration import Config

config = Config(enabled=False)


@config.spy
def foo():
    assert config.pp(1 + 2) == 3
    assert config.pp(1 + 3, 6) == (4, 6)

def main():
    call_id = eye._last_call_id
    foo()
    assert call_id is eye._last_call_id



import snoop

# Testing that a single Tracer can handle many mixed uses
snoop = snoop()


def foo(x):
    if x == 0:
        bar1(x)
        qux()
        return

    with snoop:
        # There should be line entries for these three lines,
        # no line entries for anything else in this function,
        # but calls to all bar functions should be traced
        foo(x - 1)
        bar2(x)
        qux()
    int(4)
    bar3(9)
    return x


@snoop
def bar1(_x):
    qux()


@snoop
def bar2(_x):
    qux()


@snoop
def bar3(_x):
    qux()


def qux():
    return 9  # not traced, mustn't show up


def gen():
    for i in [1, 2]:
        if i == 1:
            with snoop:
                yield i


def main():
    foo(2)
    list(gen())


if __name__ == '__main__':
    main()



import ast
import threading
import traceback
from collections import defaultdict
from datetime import datetime
from textwrap import dedent

import executing
import opcode
from pygments import highlight
from pygments.formatters.terminal256 import Terminal256Formatter
from pygments.lexers.python import Python3Lexer
from pygments.styles.monokai import MonokaiStyle
from functools import lru_cache
from snoop.utils import (ArgDefaultDict, FormattedValue,
                         ensure_tuple, my_cheap_repr,
                         optional_numeric_label, short_filename, try_statement)


class StatementsDict(dict):
    def __init__(self, source):
        super(StatementsDict, self).__init__()
        self.source = source

    def __missing__(self, key):
        statements = self.source.statements_at_line(key)
        if len(statements) == 1:
            result = list(statements)[0]
        else:
            result = None
        self[key] = result
        return result


class Source(executing.Source):
    def __init__(self, *args, **kwargs):
        super(Source, self).__init__(*args, **kwargs)
        if self.tree and self.text:
            self.highlighted = ArgDefaultDict(
                lambda style: raw_highlight(self.text, style).splitlines()
            )
        else:
            self.lines = defaultdict(lambda: u'SOURCE IS UNAVAILABLE')
            self.highlighted = defaultdict(lambda: self.lines)
        self.statements = StatementsDict(self)
        self.nodes = []
        if self.tree:
            self.tree._depth = 0
            for node in ast.walk(self.tree):
                node._tree_index = len(self.nodes)
                self.nodes.append(node)
                for child in ast.iter_child_nodes(node):
                    child._depth = node._depth + 1

    def get_text_with_indentation(self, node):
        result = self.asttokens().get_text(node)

        if not result:
            if isinstance(node, FormattedValue):
                fvals = [
                    n for n in node.parent.values
                    if isinstance(n, FormattedValue)
                ]
                return u'<f-string value{}>'.format(
                    optional_numeric_label(
                        fvals.index(node),
                        fvals,
                    )
                )
            else:
                return "<unknown>"

        if '\n' in result:
            result = ' ' * node.first_token.start[1] + result
            result = dedent(result)
        else:
            result = result.strip()
        return result


lexer = Python3Lexer(stripnl=False)


class ForceWhiteTerminal256Formatter(Terminal256Formatter):
    def _closest_color(self, r, g, b):
        result = super(ForceWhiteTerminal256Formatter, self)._closest_color(r, g, b)
        if result == 15:
            result += 6 ** 3
        return result


class NeutralMonokaiStyle(MonokaiStyle):
    """
    Monokai style that is somewhat readable on both dark and light backgrounds
    by replacing 'white' styles with no style so that the terminal can automatically
    use the appropriate foreground color.
    """
    styles = {
        k: '' if v == '#f8f8f2' else v
        for k, v in MonokaiStyle.styles.items()
    }


formatters = ArgDefaultDict(lambda style: ForceWhiteTerminal256Formatter(style=style))


def raw_highlight(code, style):
    return highlight(code, lexer, formatters[style])


cached_highlight = lru_cache(maxsize=1024)(raw_highlight)


class Event(object):
    def __init__(self, frame_info, event, arg, depth, line_no=None):
        self.frame_info = frame_info
        self.frame = frame = frame_info.frame
        self.source = frame_info.source
        self.last_line_no = frame_info.last_line_no
        self.comprehension_type = frame_info.comprehension_type

        self.event = event
        self.arg = arg
        self.depth = depth

        self.variables = []
        if line_no is None:
            line_no = frame.f_lineno
        self.line_no = line_no
        self.code = frame.f_code

        if self.event == 'call' and self.source_line.lstrip().startswith('@'):
            # If a function decorator is found, skip lines until an actual
            # function definition is found.
            while True:
                self.line_no += 1
                try:
                    if self.source_line.lstrip().startswith('def'):
                        break
                except IndexError:
                    self.line_no = self.frame.lineno
                    break

    @property
    def source_line(self):
        return self.source.lines[self.line_no - 1]

    def code_qualname(self):
        return self.source.code_qualname(self.code)

    def opname_at(self, index):
        code_byte = self.code.co_code[index]
        if not isinstance(code_byte, int):
            code_byte = ord(code_byte)
        return opcode.opname[code_byte]

    @property
    def opname(self):
        return self.opname_at(self.frame.f_lasti)

    @property
    def is_yield_value(self):
        for i in range(self.frame.f_lasti, -1, -1):
            opname = self.opname_at(i)
            if opname not in ('RESUME', 'CACHE', 'POP_TOP', 'PUSH_NULL'):
                return opname == 'YIELD_VALUE'


class DefaultFormatter(object):
    datetime_format = None

    def __init__(self, prefix, columns, color):
        prefix = str(prefix)
        if prefix and prefix == prefix.rstrip():
            prefix += u' '
        self.prefix = prefix
        self.columns = [
            column if callable(column) else
            getattr(self, u'{}_column'.format(column))
            for column in ensure_tuple(columns, split=True)
        ]
        self.column_widths = dict.fromkeys(self.columns, 0)
        if color is True:
            color = NeutralMonokaiStyle
        if color:
            self.c = Colors
            self.c.grey = formatters[color].style_string["Token.Comment"][0]

            def highlighted(code):
                return cached_highlight(code, color)

            def highlighted_source_line(event):
                return event.source.highlighted[color][event.line_no - 1]
        else:
            self.c = NoColors()

            def highlighted(code):
                return code

            def highlighted_source_line(event):
                return event.source_line

        self.highlighted = highlighted
        self.highlighted_source_line = highlighted_source_line

    def thread_column(self, _event):
        return threading.current_thread().name

    def thread_ident_column(self, _event):
        return threading.current_thread().ident

    def time_column(self, _event):
        datetime_format = self.datetime_format or '%H:%M:%S.%f'
        result = datetime.now().strftime(datetime_format)
        if self.datetime_format is None:
            result = result[:-4]
        return result

    def file_column(self, event):
        return short_filename(event.code)

    def full_file_column(self, event):
        return _get_filename(event)

    def function_column(self, event):
        return event.code.co_name

    def function_qualname_column(self, event):
        return event.code_qualname()

    def full_prefix(self, event):
        return u'{c.grey}{self.prefix}{indent}{columns} {c.reset}'.format(
            c=self.c,
            self=self,
            indent=event.depth * u'    ',
            columns=self.columns_string(event),
        )

    def format(self, event):
        # type: (Event) -> str
        lines = []

        if event.event in ('call', 'enter'):
            lines += self.format_start(event)

        statements = event.source.statements
        this_statement = statements[event.line_no]
        last_statement = statements[event.last_line_no]
        statement_start_lines = self.get_statement_start_lines(event, this_statement, last_statement)

        if event.event != 'exception':
            lines += self.format_variables(event, last_statement)

        if event.event == 'return':
            if not event.frame_info.is_ipython_cell:
                lines += self.format_return(event)
        elif event.event == 'exception':
            lines += self.format_exception(event)
        elif event.event == 'enter':
            pass
        elif event.event == 'exit':
            lines += [u'{c.green}<<< Exit with block in {func}{c.reset}'.format(
                c=self.c,
                func=event.code_qualname(),
            )]
        else:
            if not (
                    event.comprehension_type and event.event == 'line' or
                    event.frame_info.is_ipython_cell and event.event == 'call'
            ):
                lines += statement_start_lines + [self.format_event(event)]

        return self.format_lines(event, lines)

    def format_exception(self, event):
        lines = []
        exception_string = u''.join(traceback.format_exception_only(*event.arg[:2]))
        lines += [
            u'{c.red}!!! {line}{c.reset}'.format(
                c=self.c,
                line=line,
            )
            for line in exception_string.splitlines()
        ]
        lines += self.format_executing_node_exception(event)
        return lines

    def format_return(self, event):
        # If a call ends due to an exception, we still get a 'return' event
        # with arg = None. This seems to be the only way to tell the difference
        # https://stackoverflow.com/a/12800909/2482744
        opname = event.opname
        arg = event.arg
        if arg is None:
            if opname == 'END_FINALLY':
                if event.frame_info.had_exception:
                    return [u'{c.red}??? Call either returned None or ended by exception{c.reset}'
                                .format(c=self.c)]
            elif opname not in ('RETURN_VALUE', 'YIELD_VALUE', 'RETURN_CONST'):
                return [u'{c.red}!!! Call ended by exception{c.reset}'.format(c=self.c)]

        value = self.highlighted(my_cheap_repr(arg))
        if event.comprehension_type:
            prefix = plain_prefix = u'Result: '
        else:
            plain_prefix = u'<<< {description} value from {func}: '.format(
                description='Yield' if opname in ('YIELD_VALUE', 'RESUME') else 'Return',
                func=event.code_qualname(),
            )
            prefix = u'{c.green}{}{c.reset}'.format(
                plain_prefix,
                c=self.c,
            )
        return indented_lines(prefix, value, plain_prefix=plain_prefix)

    def get_statement_start_lines(self, event, this_statement, last_statement):
        result = []
        if (
                event.event != 'call' and
                this_statement and last_statement and
                this_statement != last_statement and
                this_statement.lineno != event.line_no and
                not isinstance(this_statement, try_statement)
        ):
            original_line_no = event.line_no
            for n in range(this_statement.lineno, original_line_no):
                event.line_no = n
                result.append(self.format_event(event))
            event.line_no = original_line_no
        return result

    def format_variables(self, event, last_statement):
        if last_statement:
            last_source_line = event.source.lines[last_statement.lineno - 1]
            dots = (get_leading_spaces(last_source_line)
                    .replace(' ', '.')
                    .replace('\t', '....'))
        else:
            dots = ''
            last_source_line = ''
        lines = []
        for var in event.variables:
            if (u'{} = {}'.format(*var) != last_source_line.strip()
                    and not (
                            isinstance(last_statement, ast.FunctionDef)
                            and not last_statement.decorator_list
                            and var[0] == last_statement.name
                    )
            ):
                lines += self.format_variable(var, dots, event.comprehension_type)
        return lines

    def format_start(self, event):
        if event.frame_info.is_ipython_cell:
            return []
        if event.comprehension_type:
            return [u'{type}:'.format(type=event.comprehension_type)]
        else:
            if event.event == 'enter':
                description = 'Enter with block in'
            else:
                assert event.event == 'call'
                if event.frame_info.is_generator:
                    if event.is_yield_value:
                        description = 'Re-enter generator'
                    else:
                        description = 'Start generator'
                else:
                    description = 'Call to'
            return [
                u'{c.cyan}>>> {description} {name} in File "{filename}", line {lineno}{c.reset}'.format(
                    name=event.code_qualname(),
                    filename=_get_filename(event),
                    lineno=event.line_no,
                    c=self.c,
                    description=description,
                )]

    def format_executing_node_exception(self, event):
        try:
            ex = Source.executing(event.frame)
            decorator = getattr(ex, "decorator", None)
            node = decorator or ex.node
            assert node

            description = {
                ast.Call: 'calling',
                ast.Subscript: 'subscripting',
                ast.Attribute: 'getting attribute',
                ast.Compare: 'comparing',
            }.get(type(node), 'evaluating')
            source = event.source.get_text_with_indentation(node)
            if decorator:
                description = 'calling decorator'
                source = '@' + source
            plain_prefix = u'!!! When {}: '.format(description)
            prefix = u'{c.red}{}{c.reset}'.format(plain_prefix, c=self.c)
            return indented_lines(
                prefix,
                source,
                plain_prefix=plain_prefix
            )
        except Exception:
            return []

    def columns_string(self, event):
        column_strings = []
        for column in self.columns:
            string = str(column(event))
            width = self.column_widths[column] = max(
                self.column_widths[column],
                len(string),
            )
            column_strings.append(string.ljust(width))
        return u' '.join(column_strings)

    def format_event(self, entry):
        return u'{c.grey}{line_no:4}{c.reset} | {source_line}'.format(
            source_line=self.highlighted_source_line(entry),
            c=self.c,
            **entry.__dict__
        )

    def format_variable(self, entry, dots, is_comprehension):
        name, value = entry
        if name.startswith('.') and name[1:].isdigit():
            description = u'Iterating over'
        elif is_comprehension:
            description = u'Values of {name}:'.format(name=name)
        else:
            description = u'{name} ='.format(name=name)
        prefix = u'......{dots} {description} '.format(
            description=description,
            dots=dots,
        )
        return indented_lines(prefix, self.highlighted(value))

    def format_line_only(self, event):
        return self.format_lines(event, [self.format_event(event)])

    def format_log(self, event):
        return self.format_lines(event, ['LOG:'])

    def format_log_value(self, event, source, value, depth):
        prefix = u'....{} '.format(depth * 4 * '.')
        plain_source_lines = indented_lines(prefix, source)
        highl_source_lines = indented_lines(prefix, self.highlighted(source))

        plain_prefix = plain_source_lines[-1] + ' = '
        highl_prefix = highl_source_lines[-1] + ' = '

        lines = highl_source_lines[:-1] + indented_lines(
            highl_prefix,
            self.highlighted(value),
            plain_prefix=plain_prefix
        )
        return self.format_lines(event, lines)

    def format_lines(self, event, lines):
        prefix = self.full_prefix(event)
        return u''.join([
            (
                    prefix
                    + line
                    + u'\n'
            )
            for line in lines
        ])


def get_leading_spaces(s):
    return s[:len(s) - len(s.lstrip())]


def _get_filename(event):
    # return "/path/to/example.py"
    return event.code.co_filename


class NoColors(object):
    def __getattr__(self, item):
        return ''


class Colors(object):
    grey = '\x1b[90m'
    red = '\x1b[31m\x1b[1m'
    green = '\x1b[32m\x1b[1m'
    cyan = '\x1b[36m\x1b[1m'
    reset = '\x1b[0m'


def indented_lines(prefix, string, plain_prefix=None):
    lines = str(string).splitlines() or ['']
    return [prefix + lines[0]] + [
        u' ' * len(plain_prefix or prefix) + line
        for line in lines[1:]
    ]



from tests.test_snoop import sample_traceback


@snoop
def main():
    x = 1
    y = 2
    try:
        pp.deep(lambda: x + y + bad() + 2)
    except:
        sample_traceback()


def bad():
    raise TypeError('bad')



import itertools
from collections.abc import Mapping, Sequence
from copy import deepcopy

from snoop.utils import ensure_tuple, my_cheap_repr, with_needed_parentheses


class BaseVariable(object):
    def __init__(self, source, exclude=()):
        self.source = source
        self.exclude = ensure_tuple(exclude)
        self.code = compile(source, '<variable>', 'eval')
        self.unambiguous_source = with_needed_parentheses(source)

    def items(self, frame):
        try:
            main_value = eval(self.code, frame.f_globals or {}, frame.f_locals)
        except Exception:
            return ()
        return self._items(main_value)

    def _items(self, key):
        raise NotImplementedError


class CommonVariable(BaseVariable):
    def _items(self, main_value):
        result = [(self.source, main_value)]
        for key in self._safe_keys(main_value):
            try:
                if key in self.exclude:
                    continue
                value = self._get_value(main_value, key)
            except Exception:
                continue
            result.append((
                '{}{}'.format(self.unambiguous_source, self._format_key(key)),
                value,
            ))
        return result

    def _safe_keys(self, main_value):
        try:
            for key in self._keys(main_value):
                yield key
        except Exception:
            pass

    def _keys(self, main_value):
        return ()

    def _format_key(self, key):
        raise NotImplementedError

    def _get_value(self, main_value, key):
        raise NotImplementedError


class Attrs(CommonVariable):
    def _keys(self, main_value):
        return itertools.chain(
            getattr(main_value, '__dict__', ()),
            getattr(main_value, '__slots__', ())
        )

    def _format_key(self, key):
        return '.' + key

    def _get_value(self, main_value, key):
        return getattr(main_value, key)


class Keys(CommonVariable):
    def _keys(self, main_value):
        return main_value.keys()

    def _format_key(self, key):
        return '[{}]'.format(my_cheap_repr(key).replace('\n', ' ').replace('\r', ' '))

    def _get_value(self, main_value, key):
        return main_value[key]


class Indices(Keys):
    _slice = slice(None)

    def _keys(self, main_value):
        return range(len(main_value))[self._slice]

    def __getitem__(self, item):
        assert isinstance(item, slice)
        result = deepcopy(self)
        result._slice = item
        return result


class Exploding(BaseVariable):
    def _items(self, main_value):
        if isinstance(main_value, Mapping):
            cls = Keys
        elif isinstance(main_value, Sequence):
            cls = Indices
        else:
            cls = Attrs

        return cls(self.source, self.exclude)._items(main_value)



from collections import OrderedDict

import snoop


class Foo(object):
    def __init__(self, x, y):
        self.x = x
        self.y = y


@snoop.snoop(watch_explode=('_d', '_point', 'lst + []'))
def main():
    _d = OrderedDict([('a', 1), ('b', 2), ('c', 'ignore')])
    _point = Foo(x=3, y=4)
    lst = [7, 8, 9]
    lst.append(10)



import snoop
from contextlib import contextmanager

@snoop.snoop()
def main():
    x = (
        [
            bar(),  # 1
            bar(),  # 2
        ]
    )

    with context(
            bar(),  # 1
            bar(),  # 2
    ):
        try:
            for _ in [
                bar(
                    bar(),  # 1
                    bar(),  # 2
                )
            ]:
                while bar(
                        bar(),  # 1
                        bar(),  # 2
                ):
                    pass
                else:
                    bar()
            else:
                bar()
                raise ValueError
        except (
                ValueError,
                TypeError,
        ):
            pass
        finally:
            bar(
                [
                    bar(),  # 1
                    bar(),  # 2
                ]
            )

    with context(
            bar(),  # 1
            bar(),  # 2
    ):
        if bar(
                bar(),  # 1
                bar(),  # 2
        ):
            pass
        elif [
            bar(
                bar(),  # 1
                bar(),  # 2
            )
        ]:
            pass

    return x


def bar(*_):
    pass


@contextmanager
def context(*_):
    yield



import sys

import pytest

import snoop

original_tracer = sys.gettrace()
original_tracer_active = lambda: (sys.gettrace() is original_tracer)


@snoop.snoop()
def f(x1):
    assert not original_tracer_active()
    _x2 = (yield x1)
    assert not original_tracer_active()
    _x3 = 'foo'
    assert not original_tracer_active()
    _x4 = (yield 2)
    assert not original_tracer_active()


def main():
    assert original_tracer_active()
    generator = f(0)
    assert original_tracer_active()
    first_item = next(generator)
    assert original_tracer_active()
    assert first_item == 0
    second_item = generator.send('blabla')
    assert original_tracer_active()
    assert second_item == 2
    with pytest.raises(StopIteration):
        generator.send('looloo')
    assert original_tracer_active()


if __name__ == '__main__':
    main()



import snoop


class Foo(object):
    def __init__(self):
        self.x = 2

    def square(self):
        self.x **= 2


@snoop.snoop(watch=(
        'foo.x',
        'io.__name__',
        'len(foo.__dict__["x"] * "abc")',
))
def main():
    foo = Foo()
    for i in range(2):
        foo.square()



from snoop.configuration import Config

snoop = Config(columns='time thread thread_ident file full_file function function_qualname').snoop


def main():
    @snoop
    def foo():
        x = 1
        y = x + 2

    foo()
